<?php
/**************************************************************************************************
* Copyright(c) 2009 Fabian Gerold, http://www.alumniserver.net
*
* This script is part of the AlumniServer project. It is free software; you can redistribute it
* and/or modify it under the terms of the GNU General Public License as published by the Free
* Software Foundation; either version 3 of the License, or (at your option) any later version.
*
* The GNU General Public License can be found at http://www.gnu.org/copyleft/gpl.html.
* A copy is found in the textfile GPL.txt
*
* This script is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
* even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
*
* This copyright notice MUST APPEAR in all copies of the script!
***************************************************************************************************/

if(!defined('ALUMNISERVER'))exit;
if(!isset($_SESSION['ADMIN_USER']))exit;

function getDbStructureFromFile($file){
	$structure=array();

	$fh=fopen($file, 'r');
	if(!file_exists($file)){
		return $structure;
	}
	$file_data=fread($fh,filesize($file));
	fclose($fh);

	$lines=explode(chr(10),$file_data);
	$line_array=array();
	foreach($lines as $ln){
		$ln=trim($ln);
		if($ln!='' && substr($ln,0,2)!='--' && substr($ln,0,1)!='#'){
			array_push($line_array,$ln);
		}
	}

	$table='';
	foreach($line_array as $line){
		if(strlen($table)==0){
			$line=str_replace('IF NOT EXISTS ','',$line);
			$parts=explode(' ',$line);
			if($parts[0]=='CREATE' && $parts[1]=='TABLE')	{
				$table=str_replace( '`', '', $parts[2]);
			}
		}
		else{
			if(substr($line,0,1)==')' && substr($line,-1)==';'){
				if(preg_match('/(COLLATE)[ ]*=[ ]*([a-zA-z0-9_-]+)/',$line,$table_collation_array)){
					$structure[$table]['extra']['COLLATE']=$table_collation_array[2];
				}
				else{
					// default collation
					$structure[$table]['extra']['COLLATE']='utf8_unicode_ci';
				}
				// remove table marker and start looking for the next "CREATE TABLE" statement
				$table='';
			}
			else{
				$line_content=preg_replace('/,$/','',$line);		// remove trailing commas
				$line_content=str_replace('`','',$line_content);	// remove ` ticks
				$line_content=str_replace('  ',' ',$line_content);	// remove double blanks
				$line_parts=explode(' ',$line_content,2);

				if( preg_match('/(PRIMARY|UNIQUE|FULLTEXT|INDEX|KEY)/',$line_parts[0]) ){	// key definition

					$search=array('/UNIQUE (INDEX|KEY)/', '/FULLTEXT (INDEX|KEY)/', '/INDEX/');
					$replace=array('UNIQUE', 'FULLTEXT', 'KEY');
					$line_content=preg_replace($search, $replace, $line_content);

					if(preg_match('/PRIMARY|UNIQUE|FULLTEXT/', $line_parts[0])){
						$line_parts[1]=preg_replace('/^(KEY|INDEX) /', '', $line_parts[1]);
					}

					$newParts=explode(' ',$line_parts[1],2);
					$key=$line_parts[0]=='PRIMARY' ? $line_parts[0] : $newParts[0];
					$structure[$table]['keys'][$key]=$line_content;

				}
				else{	// field definition
					// if auto_increment, remove any default values
					if(stristr($line_parts[1],'auto_increment')){
						$line_parts[1]=preg_replace('/ DEFAULT \'0\'/i','',$line_parts[1]);
					}

					// change order of "default" and "null" statements
					$line_parts[1]=preg_replace('/(.*) (DEFAULT .*) (NOT NULL)/','$1 $3 $2',$line_parts[1]);
					$line_parts[1]=preg_replace('/(.*) (DEFAULT .*) (NULL)/','$1 $3 $2',$line_parts[1]);

					$field_name=$line_parts[0];
					$structure[$table]['fields'][$field_name]=$line_parts[1];
				}
			}
		}
	}

	// remove double whitespaces
	$structure=str_replace('  ',' ',$structure);
	return $structure;
}


function getDbStructure(){
	global $dbh,$config;

	$structure=array();
	$temp_keys=array();
	$temp_keys_prefix=array();

	// get structure from existing database
	$tables=array();
	$tables_result=mysql_query('SHOW TABLE STATUS FROM `'.$config['db_db'].'`',$dbh);
	if(mysql_error()){
		return;
	}
	while($table_status=mysql_fetch_assoc($tables_result)){
		$table_name=$table_status['Name'];
		$tables[$table_name]=$table_status;

		// get fields from current table
		$columns_result=mysql_query('SHOW FULL COLUMNS FROM `'.$table_name.'`',$dbh);
		while($column=mysql_fetch_assoc($columns_result)){

			// 	Field	Type				Null	Key		Default		Extra
			//	id		int(11) unsigned	NO		PRI					auto_increment
			// 	content	text				YES	 			NULL

			$field_name=$column['Field'];
			$field=array($column['Type']);

			// collation
			if(isset($column['Collation'])){
				$field[]='COLLATE '.trim($column['Collation']);
			}

			if($column['Null']=='NO'){
				$field[]='NOT NULL';
			}

			// default value
			if(!strstr($column['Type'],'blob') && !strstr($column['Type'],'text')){
				// if not auto_increment, add default value
				if(!stristr($column['Extra'],'auto_increment')){
					if($column['Default']==''){
						if($column['Null']=='NO'){
							$field[]='DEFAULT \'\'';
						}
						else{
							$field[]='DEFAULT NULL';
						}
					}
					else{
						$field[]=trim('DEFAULT \''.addslashes($column['Default']).'\'');
					}
				}
			}

			if(isset($column['Extra'])){
				$field[]=trim(strtoupper($column['Extra']));
			}
			if(isset($column['Comment'])){
				if($column['Comment']!=''){
					$field[]='COMMENT \''.addslashes($column['Comment']).'\'';
				}
			}
			$field=trim(implode(' ',$field));
			$structure[$table_name]['fields'][$field_name]=str_replace('  ',' ',$field);
		}

		$key_result=mysql_query('SHOW KEYS FROM `'.$table_name.'`', $dbh);
		while($table_keys=mysql_fetch_assoc($key_result))	{
			$key_name=$table_keys['Key_name'];
			$col_name=$table_keys['Column_name'];
			if(isset($table_keys['Sub_part'])){
				$col_name.= '('.$table_keys['Sub_part'].')';
			}
			$temp_keys[$table_name][$key_name][$table_keys['Seq_in_index']]=$col_name;
			if($key_name=='PRIMARY'){
				$prefix='PRIMARY KEY';
			}
			else{
				if($table_keys['Index_type']=='FULLTEXT'){
					$prefix='FULLTEXT';
				}
				elseif(isset($table_keys['Non_unique'])){
					$prefix='KEY';
				}
				else{
					$prefix='UNIQUE';
				}
				$prefix.= ' '.$key_name;
			}
			$temp_keys_prefix[$table_name][$key_name]=$prefix;
		}

		// table status (collaction, etc.)
		if(isset($table_status['Collation'])){
			$structure[$table_name]['extra']['COLLATE']=$table_status['Collation'];
		}
	}

	// key information
	if(count($temp_keys)>0){
		foreach(array_keys($temp_keys) as $table){
			$key_inf=$temp_keys[$table];
			foreach(array_keys($key_inf) as $key_name){
				$index=$key_inf[$key_name];
				ksort($index);
				$structure[$table]['keys'][$key_name]=$temp_keys_prefix[$table][$key_name].' ('.implode(',',$index).')';
			}
		}
	}
	$structure=str_replace('  ',' ',$structure);
	return $structure;
}

function getExtraDiff($new_structure,$existing_structure){
	$extra=array();
	$diff=array();

	if(is_array($new_structure)){
		foreach(array_keys($new_structure) as $table){
			if(!isset($existing_structure[$table])){
				// complete table does not exist yet in db
				$extra[$table]=$new_structure[$table];
				$extra[$table]['complete_table']=1;
			}
			else{
				$key_types=array('extra','fields','keys');
				foreach($key_types as $key){
					if(!is_array($new_structure[$table][$key])){
						continue;
					}

					foreach(array_keys($new_structure[$table][$key]) as $field){
						$field=str_replace('`','',$field);
						if($field=='COLLATE'){
							continue;
						}
						$field_content=$new_structure[$table][$key][$field];

						if(!isset($existing_structure[$table][$key][$field])){
							$extra[$table][$key][$field]=$field_content;
						}
						else{
							$field_content=trim($field_content);
							$field_content=str_replace(' NOT NULL','',$field_content);
							$existing_structure[$table][$key][$field]=str_replace(' NOT NULL','',$existing_structure[$table][$key][$field]);

							if($field_content !== $existing_structure[$table][$key][$field]){
								$diff[$table][$key][$field]=$field_content;
							}
						}
					}
				}
			}
		}
	}

	$output=array(
		'extra' => $extra,
		'diff' => $diff
	);

	return $output;
}

function getUpdateQueries($extra_diff){
	global $sql;

	$statements=array();

	$keys=array('extra','diff');
	foreach( $keys as $pri_key) {
		if(!is_array($extra_diff[$pri_key])){
			continue;
		}

		foreach(array_keys($extra_diff[$pri_key]) as $table ) {
			$table_info=$extra_diff[$pri_key][$table];

			$complete_table=array();
			if(is_array($table_info['fields'])){
				foreach(array_keys($table_info['fields']) as $field_name){
					$field_value=$table_info['fields'][$field_name];
					if($table_info['complete_table']){
						$complete_table[]=$field_name.' '.$field_value;
					}
					else{
						if($pri_key=='extra'){
							$statement='ALTER TABLE `'.$table.'` ADD `'.$field_name.'` '.$field_value;
							$statements['add'][md5($statement)]=$statement;
						}
						elseif($pri_key=='diff'){
							//if(substr($field_value,0,7)=='COLLATE'){

								//if(isset($tableStatus['Collation'])){
							//}

							$statement='ALTER TABLE `'.$table.'` CHANGE `'.$field_name.'` `'.$field_name.'` '.$field_value;
							$statements['change'][md5($statement)]=$statement;
						}
					}
				}
			}
			if(is_array($table_info['keys'])){
				foreach(array_keys($table_info['keys']) as $field_name){
					$field_value=$table_info['keys'][$field_name];
					if($table_info['complete_table']){
						$complete_table[]=$field_value;
					}
					else {
						if($pri_key=='extra'){
							$statement='ALTER TABLE `'.$table.'` ADD '.$field_value;
							$statements['add'][md5($statement)]=$statement;
						}
						elseif($pri_key=='diff'){
							$statement='ALTER TABLE `'.$table.'`'.($field_name=='PRIMARY' ? ' DROP PRIMARY KEY' : ' DROP KEY '.$field_name);
							$statements['change'][md5($statement)]=$statement;
							$statement='ALTER TABLE `'.$table.'` ADD '.$field_value;
							$statements['change'][md5($statement)]=$statement;
						}
					}
				}
			}
			if(is_array($table_info['extra'])){
				$extras=array();
				foreach(array_keys($table_info['extra']) as $field_name){
					$field_value=$table_info['extra'][$field_name];
					if(!$table_info['complete_table']){
						$extras[]=$field_name.'='.$field_value;
					}
				}
				if(count($extras)>0){
					$statement='ALTER TABLE `'.$table.'` '.implode(' ',$extras);
					$statements['change'][md5($statement)]=$statement;
				}
			}
			if($table_info['complete_table']){
				$statement='CREATE TABLE `'.$table."` (\n".implode(",\n",$complete_table)."\n)";
				if($table_info['extra']){
					foreach(array_keys($table_info['extra']) as $key){
						if($key=='COLLATE'){
							continue;
						}
						// add extra attributes (CHARSET etc.)
						$statement.= ' '.$key.'='.$table_info['extra'][$key];
					}
				}
				$statements['create_table'][md5($statement)]=$statement;
			}
		}
	}
	return $statements;
}
?>